# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# __all__ = ["dr14meter"]

# from dr14meter.dr14_tmeter import main
import dr14meter.dr14_tmeter


def main():
    """Entry point for the application script"""
    dr14meter.dr14_tmeter.main()



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
import time

from dr14meter.audio_math import *
from dr14meter.out_messages import *

try:
    from scipy.interpolate import interp1d
except:
    pass


class DynCompressor:

    def __init__(self):
        self.maxDB = -3.0
        self.linear_limit = 0.3

    def set_compression_modality(self, modality="medium"):
        if modality == "very_soft" or modality == "v":
            self.maxDB = -2.0
            self.linear_limit = 0.5
        elif modality == "soft" or modality == "s":
            self.maxDB = -3.0
            self.linear_limit = 0.3
        elif modality == "medium" or modality == "m":
            self.maxDB = -5.0
            self.linear_limit = 0.3
        elif modality == "hard" or modality == "h":
            self.maxDB = -7.0
            self.linear_limit = 0.3
        elif modality == "very_hard" or modality == "vh":
            self.maxDB = -9.0
            self.linear_limit = 0.2
        else:
            self.maxDB = -3.0
            self.linear_limit = 0.3

        print("%f %f " % (self.maxDB, self.linear_limit))

    def dyn_compressor(self, Y, Fs):

        time_a = time.time()

        x = numpy.linspace(-1, 1, 21)
        y = self.c_fun(x)

        int_f = interp1d(x, y, kind='cubic')

        cY = int_f(Y)

        cY[:] = normalize(cY)

        time_b = time.time()
        dr14_log_info("dyn_compressor: Clock: %2.8f" % (time_b - time_a))

        return cY

    def c_fun(self, x):

        z = self.linear_limit

        y = numpy.zeros(x.shape)

        r = numpy.abs(x) <= (z + z * 10e-6)
        y[r] = 0.0

        r2 = numpy.abs(x) > z

        a = -self.maxDB / (z - 1.0)
        b = -a * z

        y[r2] = (a * numpy.abs(x[r2]) + b)

        return x * 10.0**(y / 20.0)



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
import os
from dr14meter.audio_decoder import AudioDecoder


class AudioTrack:

    def __init__(self):
        self.Y = numpy.array([])
        self.Fs = 0
        self.channels = 0
        self.sample_width = 0
        self._ext = -1
        self._de = AudioDecoder()

    def time(self):
        return 1 / self.Fs * self.Y.shape[0]

    def get_file_ext_code(self):
        return self._ext

    def open(self, file_name):

        self.Y = numpy.array([])
        self.Fs = 0
        self.channels = 0

        if not (os.path.exists(file_name)):
            return False

        res_f = self._de.read_track_new(file_name, self)
        self._ext = self._de.get_file_ext_code()

        return res_f



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import logging
import codecs

message_file = sys.stderr
out_file = sys.stdout
err_file = sys.stderr
mode = "verbose"

logger = logging.getLogger('dr14log')


def init_log(lev=logging.DEBUG):
    global logger

    logger = logging.getLogger('dr14log')
    logger.setLevel(logging.DEBUG)
    stream_h = logging.StreamHandler()
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    stream_h.setLevel(lev)
    stream_h.setFormatter(formatter)

    logger.addHandler(stream_h)


def dr14_log_debug(message):
    global logger
    logger.debug(message)


def dr14_log_info(message):
    global logger
    logger.info(message)

########


def print_msg(string):
    global message_file
    message_file.write("%s\n" % string)


def flush_msg():
    message_file.flush()


def print_err(string):
    global err_file
    err_file.write("Error: %s \n" % string)


def flush_err():
    err_file.flush()


def print_out(string):
    global out_file
    out_file.write("%s\n" % string)


def flush_out():
    out_file.flush()


def set_verbose_msg():
    global message_file
    global mode

    if mode == "verbose":
        return

    codecs.close(message_file)

    message_file = sys.stderr


def set_quiet_msg():
    global message_file
    global mode

    if mode == "quiet":
        return

    message_file = codecs.open(os.devnull, "w",  encoding="utf-8")



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from dr14meter.audio_math import *

try:
    import matplotlib.pyplot as pyplot
    import matplotlib.mlab as mlab
    import pylab
except:
    pass


def spectrogram(Y, Fs):

    s = Y.shape

    if len(Y.shape) > 1:
        ch = s[1]
    else:
        ch = 1

    #Ym = numpy.sum( Y , 1 ) / float(ch)

    for j in numpy.arange(ch):
        pyplot.subplot(210 + j + 1)
        pylab.specgram(Y[:, j], NFFT=Fs, Fs=Fs, cmap='gnuplot2')

        mean_x = numpy.array([0, s[0] * (1 / Fs)])
        mean_x = numpy.array([mean, mean])

        pyplot.axis([0, s[0] * (1 / Fs), 0, Fs / 2.0])

        pyplot.title("Channel %d" % j)
        pyplot.xlabel('Time [sec]')
        pyplot.ylabel('Freq. [Hz]')
        pyplot.grid(True)

        pyplot.title("Channel: %d" % int(j + 1))

    pyplot.show()



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# select track.title as title , track.id as id , dr_track.iddr , dr.dr as dr
#        from track inner join DR_Track on DR_Track.idtrack = track.id
#                   inner join dr on dr.id = DR_Track.iddr ;

# select dr.dr as dr , count(dr.dr) as dr_cnt from  DR_Track join dr on
# dr.id = DR_Track.iddr group by (dr.dr) ;

import sqlite3
import threading

from dr14meter.dr14_config import get_db_path
from dr14meter.out_messages import *


unique_db_object = 0
lock_db = threading.Lock()


def my_dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class dr_database:

    def __init__(self):

        global unique_db_object
        global lock_db

        lock_db.acquire()
        if unique_db_object > 0:
            lock_db.release()
            raise Exception(
                "Error: database.dr_database.__init__ database is not unique !")

        unique_db_object = 1
        lock_db.release()

        self._insert_session = False

        self._tracks = {}
        self._albums = {}
        self._artists = {}
        self._genre = {}
        self._codec = {}
        self._dr = {}
        self._date = {}

        self._id_artist = 0
        self._id_album = 0
        self._id_track = 0
        self._id_genre = 0
        self._id_codec = 0
        self._id_dr = 0
        self._id_date = 0

        self._current_schema_db_version = 1

    def build_database(self):
        global lock_db
        lock_db.acquire()
        if self._insert_session:
            lock_db.release()
            raise Exception(
                "Error: database.build_database It's impossible to build the database during an insertion !")
        db = self.dr14_db_main_structure_v1()

        conn = sqlite3.connect(get_db_path())

        conn.executescript(db)
        conn.commit()

        self.ungrade_db()

        lock_db.release()

    def open_insert_session(self):
        global lock_db
        lock_db.acquire()

        if self._insert_session:
            lock_db.release()
            raise Exception(
                "Error: database.open_insert_session session already opened !")
        self._insert_session = True

        self._tracks = {}
        self._albums = {}
        self._artists = {}
        self._genre = {}
        self._codec = {}
        self._dr = {}
        self._date = {}

        self._id_artist = self.query("select max(Id) from Artist").pop()[0]
        self._id_artist = 0 if self._id_artist == None else self._id_artist + 1

        self._id_album = self.query("select max(Id) from Album").pop()[0]
        self._id_album = 0 if self._id_album == None else self._id_album + 1

        self._id_track = self.query("select max(Id) from track").pop()[0]
        self._id_track = 0 if self._id_track == None else self._id_track + 1

        self._id_genre = self.query("select max(Id) from Genre").pop()[0]
        self._id_genre = 0 if self._id_genre == None else self._id_genre + 1

        self._id_codec = self.query("select max(Id) from Codec").pop()[0]
        self._id_codec = 0 if self._id_codec == None else self._id_codec + 1

        self._id_dr = self.query("select max(Id) from DR").pop()[0]
        self._id_dr = 0 if self._id_dr == None else self._id_dr + 1

        self._id_date = self.query("select max(Id) from Date").pop()[0]
        self._id_date = 0 if self._id_date == None else self._id_date + 1

        lock_db.release()

    def commit_insert_session(self):
        global lock_db
        lock_db.acquire()
        if self._insert_session == False:
            lock_db.release()
            raise Exception(
                "Error: database.commit_insert_session the session has not been opened !")

        conn = sqlite3.connect(get_db_path())
        c = conn.cursor()

        for k_dr in self._dr.keys():
            c.execute("insert into DR ( Id , DR ) values ( ? , ? ) ",
                      (k_dr, self._dr[k_dr]))

        for k_artist in self._artists.keys():
            c.execute("insert into Artist ( Id , Name ) values ( ? , ? ) ",
                      (k_artist, self._artists[k_artist]))

        for k_codec in self._codec.keys():
            c.execute("insert into Codec ( Id , Name ) values ( ? , ? ) ",
                      (k_codec, self._codec[k_codec]))

        for k_genre in self._genre.keys():
            c.execute("insert into Genre ( Id , Name ) values ( ? , ? ) ",
                      (k_genre, self._genre[k_genre]))

        for k_date in self._date.keys():
            c.execute("insert into Date ( Id , Date ) values ( ? , ? ) ",
                      (k_date, self._date[k_date]))

        for k_album in self._albums.keys():
            q = "insert into Album ( Id , sha1 , title <s1> ) values ( :Id , :sha1 , :Title <s2> ) "

            if self._albums[k_album]["disk_nr"] != None:
                q = q.replace("<s1>", " , disk_nr <s1> ")
                q = q.replace("<s2>", " , :disk_nr <s2> ")

            q = q.replace("<s1>", "")
            q = q.replace("<s2>", "")

            c.execute(q, self._albums[k_album])

            c.execute(
                "insert into DR_Album ( IdDr , IdAlbum ) values ( :dr_id , :Id )",  self._albums[k_album])

            if self._albums[k_album]["artist_id"] >= 0:
                c.execute(
                    "insert into Artist_Album ( IdArtist , IdAlbum ) values ( :artist_id , :Id ) ", self._albums[k_album])

        for k_track in self._tracks.keys():

            q = """insert into Track  ( Id  ,  Title  , rms   , peak  , duration  , bit  , bitrate  , sampling_rate  , sha1  , size  <1s> ) 
                               values ( :id , :title  , :rms  , :peak , :duration , :bit , :bitrate , :sampling_rate , :sha1 , :size <2s> ) 
                """
            if self._tracks[k_track].get("track_nr", None) != None:
                q = q.replace("<1s>", " , track_nr <1s> ")
                q = q.replace("<2s>", " , :track_nr <2s> ")

            q = q.replace("<1s>", "")
            q = q.replace("<2s>", "")

            c.execute(q, self._tracks[k_track])

            c.execute(
                "insert into DR_Track ( IdDr , IdTrack ) values ( :dr_id , :id )", self._tracks[k_track])

            c.execute(
                "insert into Codec_Track ( IdCodec , IdTrack ) values ( :codec_id , :id )", self._tracks[k_track])

            if self._tracks[k_track]["genre_id"] >= 0:
                c.execute(
                    "insert into Genre_Track ( IdGenre , IdTrack ) values ( :genre_id , :id ) ", self._tracks[k_track])

            if self._tracks[k_track]["date_id"] >= 0:
                c.execute(
                    "insert into Date_Track ( IdDate , IdTrack ) values ( :date_id , :id ) ",  self._tracks[k_track])

            if self._tracks[k_track]["artist_id"] >= 0:
                c.execute(
                    "insert into Artist_Track ( IdArtist , IdTrack ) values ( :artist_id , :id ) ", self._tracks[k_track])

            if self._tracks[k_track]["album_id"] >= 0:
                c.execute(
                    "insert into Album_Track ( IdAlbum , IdTrack ) values ( :album_id , :id ) ", self._tracks[k_track])

        conn.commit()
        conn.close()

        self._insert_session = False
        lock_db.release()

    def insert_track(self, track_sha1, title,
                     dr, rms, peak, duration,
                     codec, bit, bitrate, sampling_rate,
                     album_sha1=None, artist=None,
                     genre=None, date=None, track_nr=None, size=None):

        global lock_db
        lock_db.acquire()
        if self._insert_session == False:
            lock_db.release()
            raise Exception(
                "Error: database.insert_track the insert session has not been opened !")

        q = "select Id from track where sha1 = ? "
        rq = self.query(q, (track_sha1,))

        if len(rq) > 0:
            lock_db.release()
            return rq.pop()[0]

        #print("insert: " + title )

        artist_id = -1
        if artist != None:
            q = "select Id from artist where name = ? "
            rq = self.query(q, (artist, ))
            if len(rq) == 0 and not (artist in self._artists.values()):
                artist_id = self.__insert_artist(artist)
            elif len(rq) > 0:
                artist_id = rq.pop()[0]
            else:
                artist_id = [k for (k, v) in self._artists.items() if v == artist][
                    0]

        q = "select Id from codec where name = ? "
        rq = self.query(q, (codec, ))
        if len(rq) == 0 and not (codec in self._codec.values()):
            codec_id = self.__insert_codec(codec)
        elif len(rq) > 0:
            codec_id = rq.pop()[0]
        else:
            codec_id = [k for (k, v) in self._codec.items() if v == codec][0]

        genre_id = -1
        if genre != None:
            q = "select Id from Genre where name = ? "
            rq = self.query(q, (genre,))
            if len(rq) == 0 and not (genre in self._genre.values()):
                genre_id = self.__insert_genre(genre)
            elif len(rq) > 0:
                genre_id = rq.pop()[0]
            else:
                genre_id = [k for (k, v) in self._genre.items() if v == genre][
                    0]

        date_id = -1
        if date != None:
            date_nr = int(float(date))
            q = "select Id from Date where date = ? "
            rq = self.query(q, (date_nr, ))
            if len(rq) == 0 and not (date_nr in self._date.values()):
                date_id = self.__insert_date(date_nr)
            elif len(rq) > 0:
                date_id = rq.pop()[0]
            else:
                date_id = [k for (k, v) in self._date.items()
                           if v == date_nr][0]

        q = "select Id from DR where DR = ? "
        rq = self.query(q, (dr,))
        if len(rq) == 0 and not (dr in self._dr.values()):
            dr_id = self.__insert_dr(dr)
        elif len(rq) > 0:
            dr_id = rq.pop()[0]
        else:
            dr_id = [k for (k, v) in self._dr.items() if v == dr][0]

        album_id = -1
        if album_sha1 != None:
            rq = self.query(
                "select Id from Album where sha1 = ? ", (album_sha1, ))
            if len(rq) == 0 and not album_sha1 in [k for k in self._albums.keys()]:
                album_id = None
            elif len(rq) > 0:
                album_id = rq.pop()[0]
            else:
                album_id = [self._albums[k]["Id"]
                            for k in self._albums.keys() if k == album_sha1][0]

        self._tracks[track_sha1] = {"id": self._id_track, "sha1": track_sha1, "title": title, "dr_id": dr_id,
                                    "peak": peak, "rms": rms, "duration": duration,
                                    "codec_id": codec_id, "album_sha1": album_sha1, "artist_id": artist_id,
                                    "genre_id": genre_id, "date_id": date_id, "album_id": album_id, "track_nr": track_nr,
                                    "bit": bit, "bitrate": bitrate, "sampling_rate": sampling_rate, "size": size}

        self._id_track = self._id_track + 1

        lock_db.release()

        return self._id_track - 1

    def insert_album(self, album_sha1, title, dr, disk_nr=None, artist=None):
        global lock_db
        lock_db.acquire()

        #print( album_sha1 )
        q = "select Id from Album where sha1 = ? "
        rq = self.query(q, (album_sha1, ))

        if len(rq) > 0:
            lock_db.release()
            return rq.pop()[0]

        q = "select Id from DR where DR = ? "
        rq = self.query(q, (dr,))
        if len(rq) == 0 and not (dr in self._dr.values()):
            dr_id = self.__insert_dr(dr)
        elif len(rq) > 0:
            dr_id = rq.pop()[0]
        else:
            dr_id = [k for (k, v) in self._dr.items() if v == dr][0]

        artist_id = -1
        print(artist)
        if artist != None:
            q = "select Id from artist where name = ? "
            rq = self.query(q, (artist,))
            if len(rq) == 0 and not (artist in self._artists.values()):
                artist_id = self.__insert_artist(artist)
            elif len(rq) > 0:
                artist_id = rq.pop()[0]
            else:
                artist_id = [
                    k for (k, v) in self._artists.items() if v == artist]

        self._albums[album_sha1] = {"Id": self._id_album, "sha1": album_sha1,
                                    "Title": title, "dr_id": dr_id,
                                    "disk_nr": disk_nr, "artist_id": artist_id}

        self._id_album = self._id_album + 1

        lock_db.release()

        return self._id_album - 1

    def query(self, query, t=(), dict_factory_arg=None):

        conn = sqlite3.connect(get_db_path())

        if dict_factory_arg != None:
            conn.row_factory = dict_factory_arg

        c = conn.cursor()

        c.execute(query, t)
        res_l = c.fetchall()
        c.close()

        return res_l

    @property
    def db_version(self):
        q = """
        select max( Version ) from Db_Version ;
        """
        r = self.query(q)

        return int(r[0][0])

    def is_db_valid(self):

        if not os.path.exists(get_db_path()):
            return False

        if self._pragma_schema_version() < 21:
            return False
        else:
            return True

    def _pragma_schema_version(self):
        q = """
        pragma schema_version;
        """

        r = self.query(q)

        return int(r[0][0])

    def _pragma_integrity_check(self):
        q = """
        pragma integrity_check;
        """

        r = self.query(q)

        return int(r[0][0])

    def dr14_db_main_structure_v1(self):
        db = """
        
            create table Db_Version (
                Version integer not null unique ,
                constraint OnlyPositive check ( Version > 0 )
            ) ;
            
            insert into Db_Version ( Version ) values ( 1 ) ;
        
            create table Artist (
                Id integer primary key autoincrement ,
                Name varchar(60)
            ) ;   
            create index Artist_indx on Artist ( Name ) ;
                    
                    
            create table Track (
                Id integer primary key autoincrement ,
                Title varchar(60) ,
                track_nr integer ,
                rms float ,
                peak float ,
                duration float ,
                bit integer,
                bitrate integer ,
                sampling_rate integer ,
                size integer ,
                sha1 varchar(40) not null unique
            ) ;
            create index Track_indx on Track ( sha1 ) ;
            create index Track_title_indx on Track ( Title ) ;
            
            
            create table Album (
                Id integer primary key autoincrement ,
                sha1 varchar(40) not null ,
                Title varchar(100) ,
                disk_nr integer
            ) ;
            create index Album_indx on Album ( sha1 ) ;
            create index Album_title_indx on Track ( Title ) ;
            
            insert into Album ( Id , Title , sha1 ) values ( 0 , "None" , "0000000000000000000000000000000000000000" ) ; 
            
            
            create table DR (
                Id integer primary key autoincrement ,
                DR integer not null unique 
            ) ;
            
            create table Date (
               Id integer primary key autoincrement ,
               Date integer not null unique 
            ) ;
            
            create table Codec (
                Id integer primary key autoincrement ,
                Name varchar[15] 
            ) ;
            
            insert into Codec ( Name ) values ( "wav" ) ;
            insert into Codec ( Name ) values ( "mp3" ) ;
            insert into Codec ( Name ) values ( "flac" ) ;
            insert into Codec ( Name ) values ( "mp4" ) ;
            insert into Codec ( Name ) values ( "ogg" ) ;
            insert into Codec ( Name ) values ( "ac3" ) ;
            insert into Codec ( Name ) values ( "wma" ) ;
            insert into Codec ( Name ) values ( "ape" ) ;
                        
            create table Genre (
                Id integer primary key autoincrement ,
                Name varchar(40) 
            ) ;
            
            create table DR_Track (
                IdDr integer not null ,
                IdTrack integer not null unique ,
                primary key ( IdDr , IdTrack ),
                foreign key ( IdDr ) references DR ( Id ),
                foreign key ( IdTrack ) references Track ( Id )
            ) ;
            
            create table Codec_Track (
                IdCodec integer not null ,
                IdTrack integer not null unique ,
                primary key ( IdCodec , IdTrack ),
                foreign key ( IdCodec ) references Codec ( Id ),
                foreign key ( IdTrack ) references Track ( Id )
            ) ;            
            
            create table DR_Album (
                IdDr integer not null ,
                IdAlbum integer not null unique ,
                primary key ( IdDr , IdAlbum ),
                foreign key ( IdDr ) references DR ( Id ),
                foreign key ( IdAlbum ) references Album ( Id )
            ) ;
            
            create table Genre_Track (
                IdGenre integer not null ,
                IdTrack integer not null unique ,
                primary key ( IdGenre , IdTrack ) ,
                foreign key ( IdGenre ) references Genre ( Id ) ,
                foreign key ( IdTrack ) references Track ( Id )
            ) ;
            
            create table Date_Track (
                IdDate integer not null ,
                IdTrack integer not null unique ,
                primary key ( IdDate , IdTrack ) ,
                foreign key ( IdDate ) references Date ( Id ) ,
                foreign key ( IdTrack ) references Track ( Id )
            ) ;         
            
            create table Artist_Track (
                IdArtist integer not null ,
                IdTrack integer not null unique ,
                primary key ( IdArtist , IdTrack ) ,
                foreign key ( IdArtist ) references Artist ( Id ) ,
                foreign key ( IdTrack ) references Track ( Id )
            ) ;     
            
            create table Artist_Album (
                IdArtist integer not null ,
                IdAlbum integer not null unique ,
                primary key ( IdArtist , IdAlbum ) ,
                foreign key ( IdArtist ) references Artist ( Id ) ,
                foreign key ( IdAlbum ) references Album ( Id )
            ) ;             
                        
            create table Album_Track (
                IdAlbum integer not null ,
                IdTrack integer not null ,
                primary key ( IdAlbum , IdTrack ) ,
                foreign key ( IdAlbum ) references Album ( Id ) ,
                foreign key ( IdTrack ) references Track ( Id )
            ) ;
            
        """

        return db

    def ungrade_db(self):
        if self._current_schema_db_version == self.db_version:
            return

        #
        # upgrade db here.
        #

        return

    def dump(self):
        global lock_db
        lock_db.acquire()

        conn = sqlite3.connect(get_db_path())

        for line in conn.iterdump():
            print_out('%s\n' % line)

        lock_db.release()

    # privates methods:
    def __insert_artist(self, name):

        self._artists[self._id_artist] = name
        self._id_artist = self._id_artist + 1

        return self._id_artist - 1

    def __insert_codec(self, codec):

        self._codec[self._id_codec] = codec
        self._id_codec = self._id_codec + 1

        return self._id_codec - 1

    def __insert_genre(self, genre):

        self._genre[self._id_genre] = genre
        self._id_genre = self._id_genre + 1

        return self._id_genre - 1

    def __insert_dr(self, dr):

        self._dr[self._id_dr] = dr
        self._id_dr = self._id_dr + 1

        return self._id_dr - 1

    def __insert_date(self, date):

        self._date[self._id_date] = date
        self._id_date = self._id_date + 1

        return self._id_date - 1


class dr_database_singletone():
    __database_singletone = None

    def get(self):
        if dr_database_singletone.__database_singletone == None:
            dr_database_singletone.__database_singletone = dr_database()

        return dr_database_singletone.__database_singletone



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011 - 2012  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from dr14meter.audio_math import *

try:
    import matplotlib.pyplot as pyplot
    import matplotlib.mlab as mlab
except:
    pass


def compute_lev_hist(Y, Fs, duration=None, bins=100, plot=True, title=None):

    s = Y.shape

    if len(Y.shape) > 1:
        ch = s[1]
    else:
        ch = 1

    Ym = numpy.sum(Y, 1) / float(ch)

    if plot:
        (hist, bin_edges, patches) = pyplot.hist(Ym, 100, normed=True)

        pyplot.grid(True)

        if title != None:
            hist_title = title
        else:
            hist_title = "Hystogram of levels"

        pyplot.title(r'%s' % hist_title)

        pyplot.show()
    else:
        (hist, bin_edges) = numpy.histogram(rms, bins=bins, normed=True)

    return (hist, bin_edges)



#!/usr/bin/python

# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from dr14meter.parse_args import parse_args
from dr14meter.dynamic_range_meter import DynamicRangeMeter
from dr14meter.table import *
from dr14meter.audio_analysis import *
from dr14meter.database import dr_database_singletone
from dr14meter.dr14_global import get_exe_name, dr14_version
from dr14meter.dr14_utils import *
from dr14meter.out_messages import *
from dr14meter.dr14_config import *
from dr14meter.database_utils import *

import os
import subprocess
import sys
import logging
import numpy


def main():

    options = parse_args()

    if options.version:
        print_msg(dr14_version())
        return

    init_log(logging.DEBUG)
    logging.disable(logging.INFO)

    numpy.seterr(all='ignore')

    #print( options )

    if options.enable_database:
        enable_database()
        return

    if options.disable_database:
        enable_db(False)
        print_msg("The local DR database is disabled! ")
        return

    if options.dump_database:
        db = dr_database_singletone().get()
        db.dump()
        return

    if db_is_enabled():
        db = dr_database_singletone().get()
        f = db.is_db_valid()

        if not f:
            print_err("It seems that there is some problem with the db ... ")
            fix_problematic_database()
            return

    if options.query != None:

        if not database_exists():
            print_err("Error: The database does not exist")
            print_err(f"Error: type {get_exe_name} -q for more info.")
            return

        if len(options.query) == 0:
            query_helper()
            return

        if options.query[0] not in ["help", "top", "top_alb",
                                    "worst", "worst_alb", "top_art",
                                    "hist", "evol", "codec"]:

            print_err("Error: -q invalid parameter .")
            print_err(f"Error: type {get_exe_name} -q for more info.")
            return

        table_code = database_exec_query(options)

        if table_code is not None:
            print_out(table_code)

        return

    if options.path_name != None:
        path_name = os.path.abspath(options.path_name)
    else:
        path_name = os.path.abspath('.')

    if not(os.path.exists(path_name)):
        print_msg("Error: The input directory \"%s\"  does not exist!" % path_name)
        return

    if options.out_dir and not(os.path.exists(options.out_dir)):
        print_msg(
            "Error (-o): The target directory \"%s\"  does not exist! " % options.out_dir)
        return

    if options.quiet:
        set_quiet_msg()

    # if not options.quiet and not options.skip_version_check:
    #     l_ver = TestVer()
    #     l_ver.start()

    print_msg(path_name)
    print_msg("")

    if options.recursive:
        subdirlist = list_rec_dirs(path_name)
    else:
        subdirlist = []
        subdirlist.append(path_name)

    if run_analysis_opt(options, path_name):
        return 1

    if options.scan_file:

        dr = DynamicRangeMeter()

        dr.write_to_local_db(db_is_enabled())

        r = dr.scan_file(path_name)

        if r == 1:
            print_out("")
            print_out(dr.res_list[0]['file_name'] + " :")
            print_out("DR      = %d" % dr.res_list[0]['dr14'])
            print_out("Peak dB = %.2f" % dr.res_list[0]['dB_peak'])
            print_out("Rms dB  = %.2f" % dr.res_list[0]['dB_rms'])
            return 1
        else:
            print_msg("Error: invalid audio file")
            return 0

    if options.out_dir == "":
        out_dir = None
    else:
        out_dir = options.out_dir

    if options.append and out_dir == None:
        out_dir = path_name

    if options.files_list:
        (success, clock, r) = scan_files_list(
            options.path_name, options, out_dir)
    else:
        (success, clock, r) = scan_dir_list(subdirlist, options, out_dir)

    if success:
        print_msg("Success! ")
        print_msg("Elapsed time: %2.2f sec" % clock)
    else:
        print_msg("No audio files found\n")
        print_msg(" Usage: %s [options] path_name \n\nfor more details type \n%s --help\n" %
                  (get_exe_name(), get_exe_name()))

    if sys.platform.startswith('linux') or sys.platform.startswith('darwin'):
        subprocess.call("stty sane", shell=True)

    if not success and not database_exists():
        print_msg(" ")
        print_msg(" News ... News ... News ... News ... News  !!! ")
        print_msg(" Now there is the possibility to store all results in a database")
        print_msg(" If you want to enable this database execute the command:")
        print_msg("  > %s --enable_database " % get_exe_name())
        print_msg(" for more details visit: http://dr14tmeter.sourceforge.net/index.php/DR_Database ")

    return r


if __name__ == '__main__':
    main()



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib.metadata
import threading
import subprocess

from dr14meter.out_messages import print_msg


lock_ver = threading.Lock()

ffmpeg_cmd = None


def dr14_version():
    try:
        return importlib.metadata.version('dr14meter')
    except ModuleNotFoundError:
        print_msg("The %s function require the installation of MatPlotLib")
    return 'UNDEF'


def min_dr():
    return -10000


def get_exe_name():
    return "dr14meter"


def get_name_version():
    return f'dr14meter {dr14_version()}'


def get_ffmpeg_cmd():

    global ffmpeg_cmd

    if ffmpeg_cmd == None:
        ffmpeg_f = subprocess.call(
            "type " + "ffmpeg", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0
        avconv_f = subprocess.call(
            "type " + "avconv", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0

        if ffmpeg_f:
            ffmpeg_cmd = "ffmpeg"
        elif avconv_f:
            ffmpeg_cmd = "avconv"

    return ffmpeg_cmd


def get_home_url():
    return "https://github.com/pe7ro/p7_dr14meter"


def test_matplotlib_modules(fun_name):
    try:
        import matplotlib
    except ImportError:
        print_msg(
            "The %s function require the installation of MatPlotLib: http://matplotlib.sourceforge.net" % fun_name)
        return False

    return True

def test_mutagen(fun_name):
    try:
        import mutagen
    except ImportError:
        print_msg(
            "The %s function require the installation of MatPlotLib: http://matplotlib.sourceforge.net" % fun_name)
        return False

    return True

def test_hist_modules():
    try:
        import matplotlib
    except ImportError:
        print_msg(
            "The histogram function require the installation of MatPlotLib: http://matplotlib.sourceforge.net")
        return False

    return True


def test_compress_modules():
    try:
        import scipy
    except:
        print_msg(
            "The compression function require the installation of SciPy: http://www.scipy.org/")
        return False

    return True



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from dr14meter.audio_file_reader import *


class AudioDecoder:

    def __init__(self):
        # Attention!!! do not modify the order of this list!!!!
        # It is used for computing the sha1 of the track
        self.formats = ['.flac', '.mp3', '.ogg', '.opus', '.mp4',
                        '.m4a', '.wav', '.wv', '.ape', '.ac3', '.wma', '.dsf', '.dff']

        self._ext = -1

    def get_file_ext_code(self):
        return self._ext

    def read_track_new(self, file_name, target):

        (f, ext) = os.path.splitext(file_name)
        ext = ext.lower()

        if ext not in self.formats:
            return False

        af = AudioFileReader()

        if ext == '.mp3':
            af = Mp3FileReader()
        elif ext == '.flac':
            af = FlacFileReader()
        elif ext == '.ogg':
            af = OggFileReader()
        elif ext == '.opus':
            af = OpusFileReader()
        elif ext in ['.mp4', '.m4a']:
            af = Mp4FileReader()
        elif ext == '.wav':
            af = WavFileReader()
        elif ext == '.wav':
            af = WavPackFileReader()
        elif ext == '.wv':
            af = ApeFileReader()
        elif ext == '.ape':
            af = ApeFileReader()
        elif ext == '.ac3':
            af = Ac3FileReader()
        elif ext == '.wma':
            af = WmaFileReader()
        elif ext == '.dsf':
            af = DsfFileReader()
        elif ext == '.dff':
            af = DffFileReader()
        else:
            return False

        #af = PipeFileReader()

        self._ext = self.formats.index(ext)
        ret_f = af.read_audio_file_new(file_name, target)

        return ret_f



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy

import wave


def wav_write(filename, Fs, Y):

    amplitude = 2.0**16 - 1.0

    wav_file = wave.open(filename, "w")

    s = Y.shape

    if len(Y.shape) > 1:
        nchannels = s[1]
    else:
        nchannels = 1

    sampwidth = 2
    framerate = int(Fs)

    nframes = s[0]
    comptype = "NONE"
    compname = "no comp"

    wav_file.setparams((nchannels, sampwidth, framerate,
                        nframes, comptype, compname))

    Y_s = numpy.int16((amplitude / 2.0) * Y)
    Y_s = Y_s.tostring()

    wav_file.writeframes(Y_s)

    wav_file.close()



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dr14meter.duration import StructDuration
from dr14meter.compute_dr14 import compute_dr14
from dr14meter.compute_drv import compute_DRV

import numpy


class ComputeAudio:

    def __init__(self):
        self.duration = StructDuration()
        self.Dr_lr = numpy.zeros(2)

    def compute(self, Y, Fs):
        pass

    def get_duration(self):
        return self.duration

    def get_dr_lr(self):
        return self.Dr_lr


class ComputeDR14(ComputeAudio):

    def compute(self, Y, Fs):
        return compute_dr14(Y, Fs, self.duration, self.Dr_lr)


class ComputeDRV(ComputeAudio):

    def compute(self, Y, Fs):
        return compute_DRV(Y, Fs, self.duration, self.Dr_lr)



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011 - 2012  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class StructDuration:

    def __init__(self):
        self.tm_min = 0
        self.tm_sec = 0

    def set_samples(self, samples, Fs):
        mint = samples * (1.0 / Fs) / 60.0
        self.tm_min = int(mint)
        self.tm_sec = int((mint - self.tm_min) * 60.0)

    def to_str(self):
        return str(self.tm_min) + ":%02d" % int(self.tm_sec)

    def to_float(self):
        return float(self.tm_min) + float(int(self.tm_sec)) / 100.0

    def float_to_str(self, f):
        return "%d:%02d" % (int(f), int(100 * (f - int(f))))



# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import numpy as np
from dr14meter.audio_math import *
from dr14meter.out_messages import *
from dr14meter.my_time_formatter import *

try:
    import matplotlib.pyplot as pyplot
    import matplotlib.mlab as mlab
except:
    pass


def f_utime(t):
    v = 0.25
    a = 1 / 400
    xl = 200
    if t > xl:
        return a * t + (v - a * xl)
    else:
        return v


def plot_track(Y, Fs, Plot=True, time_range=None, utime=0):

    time_a = time.time()

    s = Y.shape

    if len(Y.shape) > 1:
        ch = s[1]
    else:
        ch = 1

    ttime = (s[0] * 1 / Fs)
    if utime == 0:
        d = ttime / f_utime(ttime)
        utime = ttime / d
        #print( utime )

    Fs = int(Fs * utime)

    sec = np.floor(s[0] / Fs) + 1
    sz = int(Fs * (sec + 1))
    tm = np.floor(np.arange(sz) / Fs)

    Yc = np.zeros(sz, dtype=np.float32)

    for i in range(ch):

        ax = pyplot.subplot(210 + i + 1)

        Yc[:] = 0.0
        Yc[0:s[0]] = Y[:, 1]

        (H, xedges, yedges) = np.histogram2d(tm, Yc, bins=(sec, 500))

        mh = np.max(H, 1)
        H = (H.T * (1 / mh))

        ax.xaxis.set_major_formatter(MyTimeFormatter(utime))
        pyplot.xticks(rotation=25)

        extent = [xedges[0], xedges[-1], yedges[-1], yedges[0]]
        pyplot.imshow(H, extent=extent, interpolation='nearest',
                      aspect='auto', cmap="hot")

    pyplot.show()

    time_b = time.time()
    dr14_log_info("plot_track: Clock: %2.8f" % (time_b - time_a))
